/*
 * Copyright(c) Huawei Technologies Co.,Ltd. 2021. All right reserved.
 */

package com.example;

import com.huawei.wienerchain.SdkClient;
import com.huawei.wienerchain.WienerChainNode;
import com.huawei.wienerchain.exception.SdkException;
import com.huawei.wienerchain.message.Builder;
import com.huawei.wienerchain.message.action.EventAction;
import com.huawei.wienerchain.message.action.TxEvent;
import com.huawei.wienerchain.message.action.event.TxEventService;
import com.huawei.wienerchain.message.build.ContractRawMessage;
import com.huawei.wienerchain.proto.common.Message.RawMessage;
import com.huawei.wienerchain.proto.common.Message.Response;
import com.huawei.wienerchain.proto.common.Message.Status;
import com.huawei.wienerchain.proto.common.NetConf.ProposeTarget;
import com.huawei.wienerchain.proto.common.TransactionOuterClass.CommonTxData;
import com.huawei.wienerchain.proto.common.TransactionOuterClass.Transaction;
import com.huawei.wienerchain.proto.common.TransactionOuterClass.TxPayload;
import com.huawei.wienerchain.proto.common.TransactionOuterClass.TxResult;
import com.huawei.wienerchain.proto.nodeservice.ContractOuterClass.Invocation;

import com.google.common.base.Charsets;
import com.google.protobuf.ByteString;
import com.google.protobuf.InvalidProtocolBufferException;

import org.bouncycastle.util.encoders.Hex;

import java.util.Locale;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

/**
 * 功能描述: 执行合约相关的操作，如发送和查询交易数据
 *
 * @author author
 * @since 2021-10-28
 */
public class Contract {

    private final String contractName;

    private final String endorserNodes;

    private final String consensusNode;

    private static final int FUTURE_TIMEOUT = 20;

    private static final int LISTEN_TIMEOUT = 200;

    public Contract(String contractName, String endorserNodes, String consensusNode) {
        this.contractName = contractName;
        this.endorserNodes = endorserNodes;
        this.consensusNode = consensusNode;
    }

    // 使用合约进行交易发送。数据修改类的操作使用此方法
    public String send(SdkClient sdkClient, String func, String[] args) throws Exception {
        String[] nodes = endorserNodes.split(",");
        if (nodes.length == 0) {
            throw new SdkException("[warn] Please input at least one node");
        }
        ContractRawMessage contractRawMessage = sdkClient.getContractRawMessage();
        RawMessage rawMessage = contractRawMessage.buildInvokeRawMsg(App.ChainID, contractName, func, args);
        RawMessage[] invokeRes = new RawMessage[nodes.length];

        for (int i = 0; i < nodes.length; i++) {
            invokeRes[i] = sdkClient.getWienerChainNode(nodes[i]).getContractAction().invoke(rawMessage)
                .get(FUTURE_TIMEOUT, TimeUnit.SECONDS);
        }

        Builder.TxRawMsg txMsg = contractRawMessage.buildTxRawMsg(invokeRes);
        String txId = Hex.toHexString(txMsg.hash);
        System.out.println("发送交易哈希： " + txId);
        RawMessage resultMsg = sdkClient.getWienerChainNode(this.consensusNode).getContractAction()
            .transaction(txMsg.msg)
            .get(FUTURE_TIMEOUT, TimeUnit.SECONDS);
        Response response = Response.parseFrom(resultMsg.getPayload());
        if (response.getStatus() != Status.SUCCESS) {
            throw new SdkException("transaction response status is: " + response.getStatus());
        }

        return txId;
    }

    // 使用合约进行交易查询。数据查询类的操作使用此方法
    public String query(SdkClient sdkClient, String func, String[] args) throws Exception {
        ContractRawMessage contractRawMessage = sdkClient.getContractRawMessage();
        // 1. 合约调用消息构建
        RawMessage rawMessage = contractRawMessage.buildInvokeRawMsg(App.ChainID, contractName, func, args);
        // 2. 查询请求消息构建与发送
        RawMessage invokeResponse = sdkClient.getWienerChainNode(App.QueryNode).getContractAction().invoke(rawMessage)
            .get(FUTURE_TIMEOUT, TimeUnit.SECONDS);
        // 3. 解析查询结果
        Response response = Response.parseFrom(invokeResponse.getPayload());
        if (response.getStatus() == Status.SUCCESS) {
            Transaction transaction = Transaction.parseFrom(response.getPayload());
            TxPayload txPayload = TxPayload.parseFrom(transaction.getPayload());
            CommonTxData commonTxData = CommonTxData.parseFrom(txPayload.getData());
            String result = commonTxData.getResponse().getPayload().toString(Charsets.UTF_8);
            System.out.println("[info] query transaction success, result is: " + result);
            return result;
        }
        System.out.println("[error] query transaction failed, error is: " + response.getStatus() + ": "
            + response.getStatusInfo());
        return null;
    }

    // 私有方法，获取调用合约所产生的结果
    private RawMessage generateInvokeResponse(SdkClient sdkClient, String nodeName,
        ContractRawMessage contractRawMessage, Invocation invocation) throws Exception {
        WienerChainNode node = sdkClient.getWienerChainNode(nodeName);
        RawMessage rawMessage = contractRawMessage.getRawMessageBuilder(invocation.toByteString())
            .setType(RawMessage.Type.DIRECT).build();
        return node.getContractAction().invoke(rawMessage).get(FUTURE_TIMEOUT, TimeUnit.SECONDS);
    }
}
